@@ -26,17 +26,27 @@ from .stream import SSHClientStreamSession, SSHServerStreamSession
 from .stream import SSHReader, SSHWriter
 
 
+def _is_regular_file(file):
+    """Return if argument is a regular file or file-like object"""
+
+    try:
+        return stat.S_ISREG(os.fstat(file.fileno()).st_mode)
+    except OSError:
+        return True
+
+
 class _UnicodeReader:
     """Handle buffering partial Unicode data"""
 
-    def __init__(self, encoding):
+    def __init__(self, encoding, textmode=False):
         self._encoding = encoding
+        self._textmode = textmode
         self._partial = b''
 
     def decode(self, data):
-        """Decode received bytes into Unicode"""
+        """Decode Unicode bytes when reading from binary sources"""
 
-        if self._encoding:
+        if self._encoding and not self._textmode:
             data = self._partial + data
             self._partial = b''
 
@@ -72,13 +82,14 @@ class _UnicodeReader:
 class _UnicodeWriter:
     """Handle encoding Unicode data before writing it"""
 
-    def __init__(self, encoding):
+    def __init__(self, encoding, textmode=False):
         self._encoding = encoding
+        self._textmode = textmode
 
     def encode(self, data):
-        """Encode Unicode bytes before writing them"""
+        """Encode Unicode bytes when writing to binary targets"""
 
-        if self._encoding:
+        if self._encoding and not self._textmode:
             data = data.encode(self._encoding)
 
         return data
@@ -88,7 +99,7 @@ class _FileReader(_UnicodeReader):
     """Forward data from a file"""
 
     def __init__(self, process, file, bufsize, datatype, encoding):
-        super().__init__(encoding)
+        super().__init__(encoding, hasattr(file, 'encoding'))
 
         self._process = process
         self._file = file
@@ -130,7 +141,7 @@ class _FileWriter(_UnicodeWriter):
     """Forward data to a file"""
 
     def __init__(self, file, encoding):
-        super().__init__(encoding)
+        super().__init__(encoding, hasattr(file, 'encoding'))
 
         self._file = file
 
@@ -318,24 +329,40 @@ class ProcessError(Error):
        addition to the usual error code, reason, and language, it
        contains the following fields:
 
-         ============ ======================================= ================
+         ============ ======================================= =================
          Field        Description                             Type
-         ============ ======================================= ================
-         exit_status  The exit status returned, or -1 if an   int
+         ============ ======================================= =================
+         env          The environment the client requested    `str` or `None`
+                      to be set for the process
+         command      The command the client requested the    `str` or `None`
+                      process to execute (if any)
+         subsystem    The subsystem the client requested the  `str` or `None`
+                      process to open (if any)
+         exit_status  The exit status returned, or -1 if an   `int`
                       exit signal is sent
-         exit_signal  The exit signal sent (if any) in the    tuple or ``None``
+         exit_signal  The exit signal sent (if any) in the    `tuple` or `None`
                       form of a tuple containing the signal
-                      name, a bool for whether a core dump
+                      name, a `bool` for whether a core dump
                       occurred, a message associated with the
                       signal, and the language the message
                       was in
-         ============ ======================================= ================
+         stdout       The output sent by the process to       `str` or `bytes`
+                      stdout (if not redirected)
+         stderr       The output sent by the process to       `str` or `bytes`
+                      stderr (if not redirected)
+         ============ ======================================= =================
 
     """
 
-    def __init__(self, exit_status, exit_signal):
+    def __init__(self, env, command, subsystem, exit_status,
+                 exit_signal, stdout, stderr):
+        self.env = env
+        self.command = command
+        self.subsystem = subsystem
         self.exit_status = exit_status
         self.exit_signal = exit_signal
+        self.stdout = stdout
+        self.stderr = stderr
 
         if exit_signal:
             signal, core_dumped, msg, lang = exit_signal
@@ -357,27 +384,35 @@ class SSHCompletedProcess(Record):
        method on :class:`SSHClientConnection` when the requested command
        has finished running. It contains the following fields:
 
-         ============ ======================================= ================
+         ============ ======================================= =================
          Field        Description                             Type
-         ============ ======================================= ================
-         exit_status  The exit status returned, or -1 if an   int
+         ============ ======================================= =================
+         env          The environment the client requested    `str` or `None`
+                      to be set for the process
+         command      The command the client requested the    `str` or `None`
+                      process to execute (if any)
+         subsystem    The subsystem the client requested the  `str` or `None`
+                      process to open (if any)
+         exit_status  The exit status returned, or -1 if an   `int`
                       exit signal is sent
-         exit_signal  The exit signal sent (if any) in the    tuple or ``None``
+         exit_signal  The exit signal sent (if any) in the    `tuple` or `None`
                       form of a tuple containing the signal
-                      name, a bool for whether a core dump
+                      name, a `bool` for whether a core dump
                       occurred, a message associated with the
                       signal, and the language the message
                       was in
-         stdout       The output sent by the process to       str or bytes
+         stdout       The output sent by the process to       `str` or `bytes`
                       stdout (if not redirected)
-         stderr       The output sent by the process to       str or bytes
+         stderr       The output sent by the process to       `str` or `bytes`
                       stderr (if not redirected)
-         ============ ======================================= ================
+         ============ ======================================= =================
 
     """
 
-    __slots__ = OrderedDict((('exit_status', None), ('exit_signal', None),
-                             ('stdout', None), ('stderr', None)))
+    __slots__ = OrderedDict((('env', None), ('command', None),
+                             ('subsystem', None), ('exit_status', None),
+                             ('exit_signal', None), ('stdout', None),
+                             ('stderr', None)))
 
 
 class SSHProcess:
@@ -427,6 +462,58 @@ class SSHProcess:
 
         return self._chan
 
+    @property
+    def logger(self):
+        """The logger associated with the process"""
+
+        return self._chan.logger
+
+    @property
+    def env(self):
+        """The environment set by the client for the process
+
+           This method returns the environment set by the client
+           when the session was opened.
+
+           :returns: A dictionary containing the environment variables
+                     set by the client
+
+        """
+
+        return self._chan.get_environment()
+
+    @property
+    def command(self):
+        """The command the client requested to execute, if any
+
+           This method returns the command the client requested to
+           execute when the process was started, if any. If the client
+           did not request that a command be executed, this method
+           will return `None`.
+
+           :returns: A `str` containing the command or `None` if
+                     no command was specified
+
+        """
+
+        return self._chan.get_command()
+
+    @property
+    def subsystem(self):
+        """The subsystem the client requested to open, if any
+
+           This method returns the subsystem the client requested to
+           open when the process was started, if any. If the client
+           did not request that a subsystem be opened, this method will
+           return `None`.
+
+           :returns: A `str` containing the subsystem name or `None`
+                     if no subsystem was specified
+
+        """
+
+        return self._chan.get_subsystem()
+
     @asyncio.coroutine
     def _create_reader(self, source, bufsize, send_eof, datatype=None):
         """Create a reader to forward data to the SSH channel"""
@@ -453,18 +540,17 @@ class SSHProcess:
                 file = os.fdopen(source, 'rb', buffering=bufsize)
             elif isinstance(source, socket.socket):
                 file = os.fdopen(source.detach(), 'rb', buffering=bufsize)
-            elif hasattr(source, 'encoding'):
-                # If file provided was opened in text mode, remove that wrapper
-                file = source.buffer
             else:
                 file = source
 
-            mode = os.fstat(file.fileno()).st_mode
-
-            if stat.S_ISREG(mode):
+            if _is_regular_file(file):
                 reader = _FileReader(self, file, bufsize,
                                      datatype, self._encoding)
             else:
+                if hasattr(source, 'buffer'):
+                    # If file was opened in text mode, remove that wrapper
+                    file = source.buffer
+
                 _, reader = \
                     yield from self._loop.connect_read_pipe(pipe_factory, file)
 
@@ -502,17 +588,16 @@ class SSHProcess:
                 file = os.fdopen(target, 'wb', buffering=bufsize)
             elif isinstance(target, socket.socket):
                 file = os.fdopen(target.detach(), 'wb', buffering=bufsize)
-            elif hasattr(target, 'encoding'):
-                # If file was opened in text mode, remove that wrapper
-                file = target.buffer
             else:
                 file = target
 
-            mode = os.fstat(file.fileno()).st_mode
-
-            if stat.S_ISREG(mode):
+            if _is_regular_file(file):
                 writer = _FileWriter(file, self._encoding)
             else:
+                if hasattr(target, 'buffer'):
+                    # If file was opened in text mode, remove that wrapper
+                    file = target.buffer
+
                 _, writer = \
                     yield from self._loop.connect_write_pipe(pipe_factory,
                                                              file)
@@ -741,34 +826,33 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
            standard input, standard output, and standard error for
            the process.
 
-           The ``stdin`` argument can be any of the following:
+           The `stdin` argument can be any of the following:
 
                * An :class:`SSHReader` object
                * A file object open for read
-               * An int file descriptor open for read
+               * An `int` file descriptor open for read
                * A connected socket object
                * A string containing the name of a file or device to open
-               * ``DEVNULL`` to provide no input to standard input
-               * ``PIPE`` to interactively write standard input
+               * `DEVNULL` to provide no input to standard input
+               * `PIPE` to interactively write standard input
 
-           The ``stdout`` and ``stderr`` arguments can be any of the
-           following:
+           The `stdout` and `stderr` arguments can be any of the following:
 
                * An :class:`SSHWriter` object
                * A file object open for write
-               * An int file descriptor open for write
+               * An `int` file descriptor open for write
                * A connected socket object
                * A string containing the name of a file or device to open
-               * ``DEVNULL`` to discard standard error output
-               * ``PIPE`` to interactively read standard error output
+               * `DEVNULL` to discard standard error output
+               * `PIPE` to interactively read standard error output
 
-           The ``stderr`` argument also accepts the value ``STDOUT`` to
+           The `stderr` argument also accepts the value `STDOUT` to
            request that standard error output be delivered to stdout.
 
            File objects passed in can be associated with plain files, pipes,
            sockets, or ttys.
 
-           The default value of ``None`` means to not change redirection
+           The default value of `None` means to not change redirection
            for that stream.
 
            :param stdin:
@@ -777,12 +861,14 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
                Target to feed data from standard output to
            :param stderr:
                Target to feed data from standard error to
-           :param int bufsize:
+           :param bufsize:
                Buffer size to use when forwarding data from a file
-           :param bool send_eof:
+           :param send_eof:
                Whether or not to send EOF to the channel when redirection
-               is complete, defaulting to ``True``. If set to ``False``,
+               is complete, defaulting to `True`. If set to `False`,
                multiple sources can be sequentially fed to the channel.
+           :type bufsize: `int`
+           :type send_eof: `bool`
 
         """
 
@@ -827,9 +913,9 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
            returning a tuple of the data written to stdout and stderr.
 
            :param input:
-               Input data to feed to standard input of the process.
-               Data should be a str if encoding is set, or bytes if not.
-           :type input: str or bytes
+               Input data to feed to standard input of the process. Data
+               should be a `str` if encoding is set, or `bytes` if not.
+           :type input: `str` or `bytes`
 
            :returns: A tuple of output to stdout and stderr
 
@@ -854,14 +940,18 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
            This method changes the width and height of the terminal
            associated with this process.
 
-           :param int width:
+           :param width:
                The width of the terminal in characters
-           :param int height:
+           :param height:
                The height of the terminal in characters
-           :param int pixwidth: (optional)
+           :param pixwidth: (optional)
                The width of the terminal in pixels
-           :param int pixheight: (optional)
+           :param pixheight: (optional)
                The height of the terminal in pixels
+           :type width: `int`
+           :type height: `int`
+           :type pixwidth: `int`
+           :type pixheight: `int`
 
            :raises: :exc:`OSError` if the SSH channel is not open
 
@@ -872,8 +962,9 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
     def send_break(self, msec):
         """Send a break to the process
 
-           :param int msec:
+           :param msec:
                The duration of the break in milliseconds
+           :type msec: `int`
 
            :raises: :exc:`OSError` if the SSH channel is not open
 
@@ -884,8 +975,9 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
     def send_signal(self, signal):
         """Send a signal to the process
 
-           :param str signal:
+           :param signal:
                The signal to deliver
+           :type signal: `str`
 
            :raises: :exc:`OSError` if the SSH channel is not open
 
@@ -920,16 +1012,17 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
            the exit status or signal information and the output sent
            to stdout and stderr if those are redirected to pipes.
 
-           If the check argument is set to ``True``, a non-zero exit
+           If the check argument is set to `True`, a non-zero exit
            status from the process with trigger the :exc:`ProcessError`
            exception to be raised.
 
-           :param bool check:
+           :param check:
                Whether or not to raise an error on non-zero exit status
+           :type check: `bool`
 
            :returns: :class:`SSHCompletedProcess`
 
-           :raises: :exc:`ProcessError` if check is set to ``True``
+           :raises: :exc:`ProcessError` if check is set to `True`
                     and the process returns a non-zero exit status
 
         """
@@ -937,9 +1030,12 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
         stdout_data, stderr_data = yield from self.communicate()
 
         if check and self.exit_status:
-            raise ProcessError(self.exit_status, self.exit_signal)
+            raise ProcessError(self.env, self.command, self.subsystem,
+                               self.exit_status, self.exit_signal,
+                               stdout_data, stderr_data)
         else:
-            return SSHCompletedProcess(self.exit_status, self.exit_signal,
+            return SSHCompletedProcess(self.env, self.command, self.subsystem,
+                                       self.exit_status, self.exit_signal,
                                        stdout_data, stderr_data)
 
 
@@ -989,31 +1085,30 @@ class SSHServerProcess(SSHProcess, SSHServerStreamSession):
            standard input, standard output, and standard error for
            the process.
 
-           The ``stdin`` argument can be any of the following:
+           The `stdin` argument can be any of the following:
 
                * An :class:`SSHWriter` object
                * A file object open for write
-               * An int file descriptor open for write
+               * An `int` file descriptor open for write
                * A connected socket object
                * A string containing the name of a file or device to open
-               * ``DEVNULL`` to discard standard error output
-               * ``PIPE`` to interactively read standard error output
+               * `DEVNULL` to discard standard error output
+               * `PIPE` to interactively read standard error output
 
-           The ``stdout`` and ``stderr`` arguments can be any of the
-           following:
+           The `stdout` and `stderr` arguments can be any of the following:
 
                * An :class:`SSHReader` object
                * A file object open for read
-               * An int file descriptor open for read
+               * An `int` file descriptor open for read
                * A connected socket object
                * A string containing the name of a file or device to open
-               * ``DEVNULL`` to provide no input to standard input
-               * ``PIPE`` to interactively write standard input
+               * `DEVNULL` to provide no input to standard input
+               * `PIPE` to interactively write standard input
 
            File objects passed in can be associated with plain files, pipes,
            sockets, or ttys.
 
-           The default value of ``None`` means to not change redirection
+           The default value of `None` means to not change redirection
            for that stream.
 
            :param stdin:
@@ -1022,12 +1117,14 @@ class SSHServerProcess(SSHProcess, SSHServerStreamSession):
                Source of data to feed to standard output
            :param stderr:
                Source of data to feed to standard error
-           :param int bufsize:
+           :param bufsize:
                Buffer size to use when forwarding data from a file
-           :param bool send_eof:
+           :param send_eof:
                Whether or not to send EOF to the channel when redirection
-               is complete, defaulting to ``True``. If set to ``False``,
+               is complete, defaulting to `True`. If set to `False`,
                multiple sources can be sequentially fed to the channel.
+           :type bufsize: `int`
+           :type send_eof: `bool`
 
         """
 
@@ -1063,56 +1160,28 @@ class SSHServerProcess(SSHProcess, SSHServerStreamSession):
         yield from self.redirect(None, None, source, bufsize, send_eof)
 
     def get_environment(self):
-        """Return the environment set by the client for the process
+        """Return the environment set by the client (deprecated)"""
 
-           This method returns the environment set by the client
-           when the session was opened.
-
-           :returns: A dictionary containing the environment variables
-                     set by the client
-
-        """
-
-        return self._chan.get_environment()
+        return self.env # pragma: no cover
 
     def get_command(self):
-        """Return the command the client requested to execute, if any
+        """Return the command the client requested to execute (deprecated)"""
 
-           This method returns the command the client requested to
-           execute when the process was started, if any. If the client
-           did not request that a command be executed, this method
-           will return ``None``.
-
-           :returns: A str containing the command or ``None`` if
-                     no command was specified
-
-        """
-
-        return self._chan.get_command()
+        return self.command # pragma: no cover
 
     def get_subsystem(self):
-        """Return the subsystem the client requested to open, if any
+        """Return the subsystem the client requested to open (deprecated)"""
 
-           This method returns the subsystem the client requested to
-           open when the process was started, if any. If the client
-           did not request that a subsystem be opened, this method will
-           return ``None``.
-
-           :returns: A str containing the subsystem name or ``None``
-                     if no subsystem was specified
-
-        """
-
-        return self._chan.get_subsystem()
+        return self.subsystem # pragma: no cover
 
     def get_terminal_type(self):
         """Return the terminal type set by the client for the process
 
            This method returns the terminal type set by the client
            when the process was started. If the client didn't request
-           a pseudo-terminal, this method will return ``None``.
+           a pseudo-terminal, this method will return `None`.
 
-           :returns: A str containing the terminal type or ``None`` if
+           :returns: A `str` containing the terminal type or `None` if
                      no pseudo-terminal was requested
 
         """
@@ -1126,7 +1195,7 @@ class SSHServerProcess(SSHProcess, SSHServerStreamSession):
            by the client. If the client didn't set any terminal size
            information, all values returned will be zero.
 
-           :returns: A tuple of four integers containing the width and
+           :returns: A tuple of four `int` values containing the width and
                      height of the terminal in characters and the width
                      and height of the terminal in pixels
 
@@ -1140,14 +1209,15 @@ class SSHServerProcess(SSHProcess, SSHServerStreamSession):
            This method looks up the value of a POSIX terminal mode
            set by the client when the process was started. If the client
            didn't request a pseudo-terminal or didn't set the requested
-           TTY mode opcode, this method will return ``None``.
+           TTY mode opcode, this method will return `None`.
 
-           :param int mode:
+           :param mode:
                POSIX terminal mode taken from :ref:`POSIX terminal modes
                <PTYModes>` to look up
+           :type mode: `int`
 
-           :returns: An int containing the value of the requested
-                     POSIX terminal mode or ``None`` if the requested
+           :returns: An `int` containing the value of the requested
+                     POSIX terminal mode or `None` if the requested
                      mode was not set
 
         """
@@ -1160,8 +1230,9 @@ class SSHServerProcess(SSHProcess, SSHServerStreamSession):
            This method can be called to report an exit status for the
            process back to the client and close the channel.
 
-           :param int status:
+           :param status:
                The exit status to report to the client
+           :type status: `int`
 
         """
 
@@ -1177,14 +1248,18 @@ class SSHServerProcess(SSHProcess, SSHServerStreamSession):
            of whether or not the process dumped core. After
            reporting the signal, the channel is closed.
 
-           :param str signal:
+           :param signal:
                The signal which caused the process to exit
-           :param bool core_dumped: (optional)
+           :param core_dumped: (optional)
                Whether or not the process dumped core
-           :param str msg: (optional)
+           :param msg: (optional)
                Details about what error occurred
-           :param str lang: (optional)
+           :param lang: (optional)
                The language the error message is in
+           :type signal: `str`
+           :type core_dumped: `bool`
+           :type msg: `str`
+           :type lang: `str`
 
         """
 
